#include <compat/serial.h>

#ifdef WIN32

#include <compat/err.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>

HANDLE serial_opendevice(char *device, speed_t baud) {
    DCB dcb;
    HANDLE h;
    char *d;
    int i;

    /* convert device from "COM19" or "COM19:" to "\\.\COM19" */
    /* see... support.microsoft.com/kb/115831 */
    d = malloc(4 + strlen(device));
    if (d == NULL)
        err(1, "malloc failed");
    strcpy(d, "\\\\.\\");
    for (i = 0; device[i] != '\0'; i++)
        d[4 + i] = device[i];
    if (i > 0 && device[i - 1] == ':')
        i--;
    d[4 + i] = '\0';

    h = CreateFile(d, GENERIC_READ | GENERIC_WRITE, 0, 0, OPEN_EXISTING, 0, 0);
    free(d);
    if (h == INVALID_HANDLE_VALUE)
        errx(1, "cannot open %s", device);

    FillMemory(&dcb, sizeof(dcb), 0);
    dcb.DCBlength = sizeof(dcb);
    if (!BuildCommDCB("57600,n,8,1", &dcb))
        errx(1, "cannot set baud rate on %s", device);
    if (baud != 0)
        dcb.BaudRate = baud;
    if (!SetCommState(h, &dcb))
        errx(1, "cannot set baud rate on %s", device);

    if (!SetupComm(h, 1024, 1024))
        errx(1, "cannot set buffer size on %s", device);

    return h;
}

int serial_getc(HANDLE h, int timeout_ms) {
    COMMTIMEOUTS cmt;
    DWORD read = -1;
    unsigned char c;

    cmt.ReadIntervalTimeout = timeout_ms == -1 ? 0 : timeout_ms;
    cmt.ReadTotalTimeoutMultiplier = timeout_ms == -1 ? 0 : 1;
    cmt.ReadTotalTimeoutConstant = 0;
    cmt.WriteTotalTimeoutMultiplier = 0;
    cmt.WriteTotalTimeoutConstant = 0;

    if (!SetCommTimeouts(h, &cmt))
        errx(1, "cannot set timeouts on device");

    return ReadFile(h, &c, 1, &read, NULL) == 0 ? SERIAL_TIMEOUT
         : read == 0                            ? SERIAL_TIMEOUT
         :                                        c;
}

int serial_putc(HANDLE h, unsigned char c) {
    DWORD write = 0;
    WriteFile(h, &c, 1, &write, NULL);
    return write;
}

#else

#include <compat/err.h>
#include <errno.h>
#include <stdint.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>

int serial_opendevice(char *device, speed_t baud) {
    int fd;
    struct termios t;

    fd = open(device, O_RDWR);
    if (fd == -1)
        err(1, "cannot open %s", device);
    if (flock(fd, LOCK_EX | LOCK_NB) != 0)
        err(1, "cannot lock %s", device);
    tcgetattr(fd, &t);
    cfmakeraw(&t);
    cfsetispeed(&t, baud);
    cfsetospeed(&t, baud);
    if (tcsetattr(fd, 0, &t) == -1)
        err(1, "cannot configure %s", device);
    return fd;
}

int serial_getc(int fd, int timeout_ms) {
    int c, i;
    uint8_t cc;
    ssize_t ret;

    if (timeout_ms != -1) {
        for (i = 0; i < timeout_ms; i++) {
            ret = ioctl(fd, FIONREAD, &c);
            if (ret == -1 && errno != EAGAIN)
                return SERIAL_ERROR;
            if (c != 0)
                break;
            usleep(1000);
        }
        if (c == 0)
            return SERIAL_TIMEOUT;
    }

    for(;;) {
        ret = read(fd, &cc, 1);
        if (ret == 1)
            return cc;
        if (ret != -1 || errno != EAGAIN)
            return SERIAL_ERROR;
    }
}

int serial_putc(int fd, uint8_t c) {
    for(;;) {
        ssize_t ret = write(fd, &c, 1);
        if (ret == 1)
            return 1;
        if (ret != -1 || errno != EAGAIN)
            return 0;
    }
}

#endif

void serial_put(serial_device_t device, void *buf, int len) {
    uint8_t *c = buf;
    while (len--)
        serial_putc(device, *c++);
}
